@@ -122,7 +122,11 @@ err_http <- function(msg, resp = NULL, context = list()) {
     }
   }
   # Compute non-sensitive body digest for debugging if available
+  # and extract RFC 6749 §5.2 structured error fields when present
   body_digest <- NULL
+  oauth_error <- NULL
+  oauth_error_description <- NULL
+  oauth_error_uri <- NULL
   if (!is.null(resp) && inherits(resp, "httr2_response")) {
     bs <- try(httr2::resp_body_string(resp), silent = TRUE)
     if (!inherits(bs, "try-error")) {
@@ -130,6 +134,22 @@ err_http <- function(msg, resp = NULL, context = list()) {
       if (!inherits(dig, "try-error")) {
         body_digest <- paste0(sprintf("%02x", as.integer(dig)), collapse = "")
       }
+      # RFC 6749 §5.2: try to extract structured error fields from JSON body
+      parsed <- try(
+        jsonlite::fromJSON(bs, simplifyVector = TRUE),
+        silent = TRUE
+      )
+      if (!inherits(parsed, "try-error") && is.list(parsed)) {
+        if (is_valid_string(parsed[["error"]])) {
+          oauth_error <- parsed[["error"]]
+        }
+        if (is_valid_string(parsed[["error_description"]])) {
+          oauth_error_description <- parsed[["error_description"]]
+        }
+        if (is_valid_string(parsed[["error_uri"]])) {
+          oauth_error_uri <- parsed[["error_uri"]]
+        }
+      }
     }
   }
 
@@ -140,7 +160,10 @@ err_http <- function(msg, resp = NULL, context = list()) {
       message = msg,
       status = status,
       url = url,
-      body_digest = body_digest
+      body_digest = body_digest,
+      oauth_error = oauth_error,
+      oauth_error_description = oauth_error_description,
+      oauth_error_uri = oauth_error_uri
     ),
     context
   )
@@ -157,6 +180,21 @@ err_http <- function(msg, resp = NULL, context = list()) {
   } else {
     character()
   }
+  # RFC 6749 §5.2: surface structured error fields from token endpoint
+  oauth_error_msg <- if (!is.null(oauth_error)) {
+    reason <- oauth_error
+    if (!is.null(oauth_error_description)) {
+      reason <- paste0(reason, ": ", oauth_error_description)
+    }
+    stats::setNames(paste0("OAuth error: ", reason), "x")
+  } else {
+    character()
+  }
+  oauth_error_uri_msg <- if (!is.null(oauth_error_uri)) {
+    stats::setNames(paste0("Error URI: ", oauth_error_uri), "i")
+  } else {
+    character()
+  }
   url_msg <- if (!is.null(url)) {
     stats::setNames(paste0("URL: ", url), "i")
   } else {
@@ -172,6 +210,8 @@ err_http <- function(msg, resp = NULL, context = list()) {
     format_header("HTTP request failed"),
     bullets,
     status_msg,
+    oauth_error_msg,
+    oauth_error_uri_msg,
     url_msg,
     trace_msg,
     body_msg
@@ -184,6 +224,9 @@ err_http <- function(msg, resp = NULL, context = list()) {
     status = status,
     url = url,
     body_digest = body_digest,
+    oauth_error = oauth_error,
+    oauth_error_description = oauth_error_description,
+    oauth_error_uri = oauth_error_uri,
     context = context
   )
 }
@@ -303,12 +346,30 @@ emit_trace_event <- function(event) {
     event
   })
   if (is.function(hook)) {
-    # Best-effort, never fail the error path
-    try(hook(event), silent = TRUE)
+    # Surface hook errors as warnings so they are visible in the main process
+    # (async_dispatch captures warnings and replays them on the main thread).
+    tryCatch(
+      hook(event),
+      error = function(e) {
+        warning(
+          "[shinyOAuth] trace_hook error: ",
+          conditionMessage(e),
+          call. = FALSE
+        )
+      }
+    )
   }
   if (is.function(audit_hook)) {
-    # Allow separate audit hook consumers
-    try(audit_hook(event), silent = TRUE)
+    tryCatch(
+      audit_hook(event),
+      error = function(e) {
+        warning(
+          "[shinyOAuth] audit_hook error: ",
+          conditionMessage(e),
+          call. = FALSE
+        )
+      }
+    )
   }
   invisible(NULL)
 }
