@@ -92,7 +92,7 @@ is_ok_host <- function(
       plausible <- (grepl("^\\[", x) ||
         grepl("localhost", x, ignore.case = TRUE) ||
         grepl("^[0-9]{1,3}(\\.[0-9]{1,3}){3}", x) ||
-        grepl("::", x) ||
+        grepl("::", x, fixed = TRUE) ||
         grepl("\\.", x) ||
         grepl(":[0-9]{1,5}", x)) &&
         !grepl("\\s", x)
@@ -145,7 +145,9 @@ is_ok_host <- function(
     if (!host_matches_any(host, allowed_hosts)) {
       # Defensive fallback: direct equality for IPv6 literals against normalized patterns
       if (
-        !is.null(allowed_hosts) && length(allowed_hosts) > 0 && grepl(":", host)
+        !is.null(allowed_hosts) &&
+          length(allowed_hosts) > 0 &&
+          grepl(":", host, fixed = TRUE)
       ) {
         ah <- tolower(vapply(
           allowed_hosts,
@@ -283,18 +285,24 @@ host_glob_to_regex <- function(pat) {
 
   if (startsWith(pat, ".")) {
     core <- substr(pat, 2, nchar(pat))
+    # Escape regex metacharacters EXCEPT * and ? which are glob wildcards
     core_esc <- gsub(
-      "([.\\^$|()\\[\\]{}+?\\\\])",
+      "([.\\^$|()\\[\\]{}+\\\\])",
       "\\\\\\1",
       core,
       perl = TRUE
     )
+    # Convert glob wildcards to regex equivalents
+    core_esc <- gsub("*", ".*", core_esc, fixed = TRUE)
+    core_esc <- gsub("?", ".", core_esc, fixed = TRUE)
     return(paste0("^(?:", core_esc, "|(?:[^.]+\\.)+", core_esc, ")$"))
   }
 
-  esc <- gsub("([.\\^$|()\\[\\]{}+?\\\\])", "\\\\\\1", pat, perl = TRUE)
-  esc <- gsub("\\*", ".*", esc, perl = TRUE)
-  esc <- gsub("\\?", ".", esc, perl = TRUE)
+  # Escape regex metacharacters EXCEPT * and ? which are glob wildcards
+  esc <- gsub("([.\\^$|()\\[\\]{}+\\\\])", "\\\\\\1", pat, perl = TRUE)
+  # Convert glob wildcards to regex equivalents
+  esc <- gsub("*", ".*", esc, fixed = TRUE)
+  esc <- gsub("?", ".", esc, fixed = TRUE)
   paste0("^", esc, "$")
 }
 
@@ -326,7 +334,7 @@ host_matches_any <- function(host, patterns) {
   if (host_lc %in% patterns) {
     return(TRUE)
   }
-  host_br <- if (grepl(":", host_lc) && !grepl("^\\[", host_lc)) {
+  host_br <- if (grepl(":", host_lc, fixed = TRUE) && !grepl("^\\[", host_lc)) {
     paste0("[", host_lc, "]")
   } else {
     host_lc
@@ -393,20 +401,70 @@ parse_url_host <- function(url, label = "url") {
   sub("^\\[([^\\]]+)\\]$", "\\1", h)
 }
 
-#' Internal: Resolve issuer from discovery with strict host check
+normalize_issuer_url <- function(url, label = "issuer") {
+  parsed <- try(httr2::url_parse(url), silent = TRUE)
+
+  if (inherits(parsed, "try-error")) {
+    err_config(c(
+      "x" = sprintf("Could not parse %s", label),
+      "!" = sprintf("Value: '%s'", url)
+    ))
+  }
+
+  scheme <- tolower((parsed$scheme %||% ""))
+  host <- tolower(trimws(parsed$hostname %||% ""))
+  host <- sub("^\\[([^\\]]+)\\](?::.*)?$", "\\1", host, perl = TRUE)
+  host <- sub("\\.$", "", host)
+
+  if (!nzchar(host)) {
+    err_config(c(
+      "x" = sprintf("%s does not include a hostname", label),
+      "!" = sprintf("Value: '%s'", url)
+    ))
+  }
+
+  port <- parsed$port %||% ""
+  port <- as.character(port)
+  port <- if (nzchar(port)) paste0(":", port) else ""
+
+  path <- parsed$path %||% ""
+  path <- as.character(path)
+  path <- if (!nzchar(path) || identical(path, "/")) {
+    ""
+  } else {
+    path <- sub("/+$", "", path)
+    if (!startsWith(path, "/")) paste0("/", path) else path
+  }
+
+  query <- parsed$query %||% ""
+  query <- as.character(query)
+  query <- if (nzchar(query)) paste0("?", query) else ""
+
+  fragment <- parsed$fragment %||% ""
+  fragment <- as.character(fragment)
+  fragment <- if (nzchar(fragment)) paste0("#", fragment) else ""
+
+  paste0(scheme, "://", host, port, path, query, fragment)
+}
+
+#' Internal: Resolve issuer from discovery with issuer matching policy
 #'
-#' Prefers the discovery issuer when provided. By default, enforces that
-#' the input issuer and discovery issuer have identical scheme+host. On mismatch,
-#' raises err_config. Callers can explicitly allow host mismatch by passing
-#' `issuer_match = TRUE` enforces equality; set to FALSE to allow mismatch.
+#' Prefers the discovery issuer when provided.
+#'
+#' Matching is controlled by `issuer_match`:
+#' - "url": require full issuer URL match after trailing-slash normalization
+#' - "host": require scheme+host match only (explicit opt-out)
+#' - "none": do not validate issuer consistency
 #'
 #' @keywords internal
 #' @noRd
 validate_discovery_issuer <- function(
   issuer_input,
   issuer_discovered,
-  issuer_match = TRUE
+  issuer_match = c("url", "host", "none")
 ) {
+  issuer_match <- match.arg(issuer_match)
+
   # Prefer discovered when available; otherwise fall back to input
   iss <- issuer_discovered %||% issuer_input
 
@@ -415,27 +473,43 @@ validate_discovery_issuer <- function(
     return(iss)
   }
 
-  # If matching is not required, allow mismatch
-  if (!isTRUE(issuer_match)) {
+  if (identical(issuer_match, "none")) {
     return(iss)
   }
 
-  p_in <- parse_url_components(issuer_input, "issuer")
-  p_dc <- parse_url_components(issuer_discovered, "discovery issuer")
+  if (identical(issuer_match, "host")) {
+    p_in <- parse_url_components(issuer_input, "issuer")
+    p_dc <- parse_url_components(issuer_discovered, "discovery issuer")
 
-  if (
-    !identical(p_in$scheme, p_dc$scheme) || !identical(p_in$host, p_dc$host)
-  ) {
+    if (
+      !identical(p_in$scheme, p_dc$scheme) || !identical(p_in$host, p_dc$host)
+    ) {
+      err_config(
+        c(
+          "x" = "OIDC discovery issuer mismatch",
+          "!" = sprintf(
+            "Input '%s://%s' vs discovery '%s://%s'",
+            p_in$scheme,
+            p_in$host,
+            p_dc$scheme,
+            p_dc$host
+          )
+        )
+      )
+    }
+
+    return(iss)
+  }
+
+  in_norm <- normalize_issuer_url(issuer_input, "issuer")
+  dc_norm <- normalize_issuer_url(issuer_discovered, "discovery issuer")
+
+  if (!identical(in_norm, dc_norm)) {
     err_config(
       c(
         "x" = "OIDC discovery issuer mismatch",
-        "!" = sprintf(
-          "Input '%s://%s' vs discovery '%s://%s'",
-          p_in$scheme,
-          p_in$host,
-          p_dc$scheme,
-          p_dc$host
-        )
+        "!" = sprintf("Input '%s' vs discovery '%s'", in_norm, dc_norm),
+        "i" = "Set issuer_match = 'host' to compare only scheme+host (not recommended)"
       )
     )
   }
