@@ -135,14 +135,17 @@ test_that("handle_callback validates browser token, PKCE verifier, and nonce", {
         access_token = "at",
         expires_in = 3600,
         id_token = "dummy.jwt.token",
-        token_type = "Bearer"
+        token_type = "Bearer",
+        scope = "openid"
       )
     },
     validate_id_token = function(
       client,
       id_token,
       expected_nonce = NULL,
-      expected_sub = NULL
+      expected_sub = NULL,
+      expected_access_token = NULL,
+      max_age = NULL
     ) {
       # Pretend success if expected_nonce non-empty
       if (isTRUE(client@provider@use_nonce)) {
