@@ -59,7 +59,9 @@ test_that("login flow: get_userinfo is called after validate_id_token", {
       client,
       id_token,
       expected_nonce = NULL,
-      expected_sub = NULL
+      expected_sub = NULL,
+      expected_access_token = NULL,
+      max_age = NULL
     ) {
       call_order <<- c(call_order, "validate_id_token")
       invisible(list(sub = "user123", iss = "https://test.example.com"))
@@ -149,7 +151,9 @@ test_that("login flow: get_userinfo not called when ID token validation fails",
         client,
         id_token,
         expected_nonce = NULL,
-        expected_sub = NULL
+        expected_sub = NULL,
+        expected_access_token = NULL,
+        max_age = NULL
       ) {
         shinyOAuth:::err_id_token("Simulated ID token validation failure")
       },
@@ -245,7 +249,9 @@ test_that("verify_token_set: initial login does not call userinfo match (no user
       client,
       id_token,
       expected_nonce = NULL,
-      expected_sub = NULL
+      expected_sub = NULL,
+      expected_access_token = NULL,
+      max_age = NULL
     ) {
       invisible(list(sub = "user123"))
     },
@@ -334,7 +340,9 @@ test_that("verify_token_set: refresh with both userinfo and id_token calls match
       client,
       id_token,
       expected_nonce = NULL,
-      expected_sub = NULL
+      expected_sub = NULL,
+      expected_access_token = NULL,
+      max_age = NULL
     ) {
       invisible(list(sub = "user123"))
     },
@@ -422,7 +430,9 @@ test_that("verify_token_set: refresh without userinfo skips match", {
       client,
       id_token,
       expected_nonce = NULL,
-      expected_sub = NULL
+      expected_sub = NULL,
+      expected_access_token = NULL,
+      max_age = NULL
     ) {
       invisible(list(sub = "user123"))
     },
@@ -581,7 +591,9 @@ test_that("handle_callback: userinfo/id_token match IS performed after userinfo
       client,
       id_token,
       expected_nonce = NULL,
-      expected_sub = NULL
+      expected_sub = NULL,
+      expected_access_token = NULL,
+      max_age = NULL
     ) {
       invisible(list(sub = "user123", iss = "https://test.example.com"))
     },
@@ -676,7 +688,9 @@ test_that("handle_callback: userinfo/id_token mismatch aborts login", {
         client,
         id_token,
         expected_nonce = NULL,
-        expected_sub = NULL
+        expected_sub = NULL,
+        expected_access_token = NULL,
+        max_age = NULL
       ) {
         # ID token says sub = "user123"
         invisible(list(sub = "user123", iss = "https://test.example.com"))
